@@ -16,10 +16,8 @@ from abc import ABCMeta
 from unicodedata import name as unicode_name
 from decimal import Decimal, DecimalException
 from typing import Any, cast, overload, no_type_check_decorator, Callable, \
-    ClassVar, FrozenSet, Dict, Generic, List, Optional, Union, Tuple, Type, \
-    Pattern, Match, MutableMapping, MutableSequence, Iterator, Set, TypeVar
-
-from .datatypes import AtomicValueType
+    Dict, Generic, List, Optional, Union, Tuple, Type, Pattern, Match, \
+    MutableMapping, MutableSequence, Iterator, TypeVar
 
 #
 # Simple top-down parser based on Vaughan Pratt's algorithm (Top Down Operator Precedence).
@@ -80,8 +78,8 @@ def _symbol_to_classname(symbol: str) -> str:
 
 class MultiLabel:
     """
-    Helper class for defining multi-value label for tokens. Useful when a symbol has more roles.
-    A label of this type has equivalence with each of its values.
+    Helper class for defining multi-value label for tokens. Useful when a symbol
+    has more roles. A label of this type has equivalence with each of its values.
 
     Example:
         label = MultiLabel('function', 'operator')
@@ -142,7 +140,7 @@ class Token(MutableSequence[TK]):
     :cvar lbp: Pratt's left binding power, defaults to 0.
     :cvar rbp: Pratt's right binding power, defaults to 0.
     :cvar pattern: the regex pattern used for the token class. Defaults to the \
-    escaped symbol. Can be customized to match more detailed conditions (eg. a \
+    escaped symbol. Can be customized to match more detailed conditions (e.g. a \
     function with its left round bracket), in order to simplify the related code.
     :cvar label: defines the typology of the token class. Its value is used in \
     representations of the token instance and can be used to restrict code choices \
@@ -151,36 +149,35 @@ class Token(MutableSequence[TK]):
     the XPath parsers). In the base parser class defaults to 'symbol' with 'literal' \
     and 'operator' as possible alternatives. If set by a tuple of values the token \
     class label is transformed to a multi-value label, that means the token class can \
-    covers multiple roles (eg. as XPath function or axis). In those cases the definitive \
+    covers multiple roles (e.g. as XPath function or axis). In those cases the definitive \
     role is defined at parse time (nud and/or led methods) after the token instance creation.
     """
     lbp: int = 0           # left binding power
     rbp: int = 0           # right binding power
     symbol: str = ''       # the token identifier
+    lookup_name: str = ''  # the key in symbol table, usually matches the symbol.
     label: str = 'symbol'  # optional label
     pattern: Optional[str] = None  # a custom regex pattern for building the tokenizer
 
-    __slots__ = '_items', 'parser', 'value', '_source', 'span'
+    __slots__ = '_items', 'parser', 'value', 'span'
 
     _items: List[TK]
     parser: 'Parser[Token[TK]]'
-    value: Optional[AtomicValueType]
-    _source: str
+    value: Optional[Any]
     span: Tuple[int, int]
 
     def __init__(self, parser: 'Parser[Token[TK]]',
-                 value: Optional[AtomicValueType] = None) -> None:
+                 value: Optional[Any] = None) -> None:
         self._items = []
         self.parser = parser
         self.value = value if value is not None else self.symbol
-        self._source = parser.source
         self.span = (0, 0) if parser.next_match is None else parser.next_match.span()
 
     @overload
-    def __getitem__(self, i: int) -> TK: ...
+    def __getitem__(self, i: int) -> TK: ...  # pragma: no cover
 
     @overload
-    def __getitem__(self, s: slice) -> MutableSequence[TK]: ...
+    def __getitem__(self, s: slice) -> MutableSequence[TK]: ...  # pragma: no cover
 
     def __getitem__(self, i: Union[int, slice]) \
             -> Union[TK, MutableSequence[TK]]:
@@ -205,11 +202,9 @@ class Token(MutableSequence[TK]):
             return '%r %s' % (self.symbol, str(self.label))
 
     def __repr__(self) -> str:
-        symbol, value = self.symbol, self.value
-        if value != symbol:
-            return '%s(value=%r)' % (self.__class__.__name__, value)
-        else:
-            return '%s()' % self.__class__.__name__
+        if self.value == self.symbol:
+            return '%s(%r)' % (self.__class__.__name__, self.parser)
+        return '%s(%r, %r)' % (self.__class__.__name__, self.parser, self.value)
 
     def __eq__(self, other: object) -> bool:
         if isinstance(other, Token):
@@ -228,15 +223,13 @@ class Token(MutableSequence[TK]):
         elif self.symbol in SPECIAL_SYMBOLS:
             return '(%r)' % self.value
         elif self.symbol == '(':
-            if not self:
-                return '()'
-            elif len(self) == 1:
+            if len(self) == 1:
                 return self[0].tree
-            return '(%s)' % ' '.join(item.tree for item in self)
+            return f"({' '.join(item.tree for item in self)})"
         elif not self:
             return '(%s)' % self.symbol
         else:
-            return '(%s %s)' % (self.symbol, ' '.join(item.tree for item in self))
+            return f"({self.symbol} {' '.join(item.tree for item in self)})"
 
     @property
     def source(self) -> str:
@@ -247,7 +240,7 @@ class Token(MutableSequence[TK]):
         elif symbol == '(decimal)':
             return str(self.value)
         elif symbol in SPECIAL_SYMBOLS:
-            return repr(self.value)
+            return repr(self.value).replace(r'\\', '\\')
         else:
             length = len(self)
             if not length:
@@ -265,10 +258,55 @@ class Token(MutableSequence[TK]):
     def position(self) -> Tuple[int, int]:
         """A tuple with the position of the token in terms of line and column."""
         token_index = self.span[0]
-        line = self._source[:token_index].count('\n') + 1
+        line = self.parser.source[:token_index].count('\n') + 1
         if line == 1:
             return 1, token_index + 1
-        return line, token_index - self._source[:token_index].rindex('\n')
+        return line, token_index - self.parser.source[:token_index].rindex('\n')
+
+    def as_name(self) -> 'Token[TK]':
+        """Returns a new '(name)' token for resolving ambiguous states."""
+        assert self.parser.name_pattern.match(self.symbol) is not None, \
+            "Token symbol is not compatible with the name pattern!"
+
+        token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+        token.span = self.span
+        return token
+
+    def is_source_start(self) -> bool:
+        """
+        Returns `True` if the token is positioned at the start
+        of the source, ignoring the spaces.
+        """
+        return not bool(self.parser.source[0:self.span[0]].strip())
+
+    def is_line_start(self) -> bool:
+        """
+        Returns `True` if the token is positioned at the start
+        of a source line, ignoring the spaces.
+        """
+        token_index = self.span[0]
+        try:
+            line_start = self.parser.source[:token_index].rindex('\n') + 1
+        except ValueError:
+            return not bool(self.parser.source[:token_index].strip())
+        else:
+            return not bool(self.parser.source[line_start:token_index].strip())
+
+    def is_spaced(self, before: bool = True, after: bool = True) -> bool:
+        """
+        Returns `True` if the token has extra spaces (whitespace, tab or newline)
+        immediately before or after it.
+
+        :param before: if `True` considers also the extra spaces before the token.
+        :param after: if `True` considers also the extra spaces after the token.
+        """
+        start, end = self.span
+        try:
+            if before and start > 0 and self.parser.source[start - 1] in ' \t\n':
+                return True
+            return after and self.parser.source[end] in ' \t\n'
+        except IndexError:
+            return False
 
     def nud(self) -> TK:
         """Pratt's null denotation method"""
@@ -290,22 +328,7 @@ class Token(MutableSequence[TK]):
         tk: 'Token[TK]'
 
         while True:
-            try:
-                tk = next(children)
-            except StopIteration:
-                try:
-                    parent, children = status.pop()
-                except IndexError:
-                    if parent is not None:
-                        if not symbols or parent.symbol in symbols:
-                            yield parent
-                    return
-                else:
-                    if parent is not None:
-                        if not symbols or parent.symbol in symbols:
-                            yield parent
-                        parent = None
-            else:
+            for tk in children:
                 if parent is not None and len(parent._items) == 1:
                     if not symbols or parent.symbol in symbols:
                         yield parent
@@ -321,6 +344,20 @@ class Token(MutableSequence[TK]):
                     continue
                 status.append((parent, children))
                 parent, children = tk, iter(tk)
+                break
+            else:
+                try:
+                    parent, children = status.pop()
+                except IndexError:
+                    if parent is not None:
+                        if not symbols or parent.symbol in symbols:
+                            yield parent
+                    return
+                else:
+                    if parent is not None:
+                        if not symbols or parent.symbol in symbols:
+                            yield parent
+                        parent = None
 
     def expected(self, *symbols: str, message: Optional[str] = None) -> None:
         if symbols and self.symbol not in symbols:
@@ -361,7 +398,6 @@ class ParserMeta(ABCMeta):
     literals_pattern: Pattern[str]
     name_pattern: Pattern[str]
     tokenizer: Optional[Pattern[str]]
-    SYMBOLS: Set[str]
     symbol_table: Dict[str, Token[Any]]
 
     def __new__(mcs, name: str, bases: Tuple[Type[Any], ...], namespace: Dict[str, Any]) \
@@ -384,12 +420,6 @@ class ParserMeta(ABCMeta):
             cls.name_pattern = re.compile(r'[A-Za-z0-9_]+')
         if 'tokenizer' not in namespace:
             cls.tokenizer = None
-        if 'SYMBOLS' not in namespace:
-            cls.SYMBOLS = set()
-            for base_class in bases:
-                if hasattr(base_class, 'SYMBOLS'):
-                    cls.SYMBOLS.update(base_class.SYMBOLS)
-                    break
         if 'symbol_table' not in namespace:
             cls.symbol_table = {}
             for base_class in bases:
@@ -404,25 +434,21 @@ TK_co = TypeVar('TK_co', bound=Token[Any], covariant=True)
 
 class Parser(Generic[TK_co], metaclass=ParserMeta):
     """
-    Parser class for implementing a Top Down Operator Precedence parser.
+    Parser class for implementing a Top-Down Operator Precedence parser.
 
-    :cvar SYMBOLS: the symbols of the definable tokens for the parser. In the base class it's an \
-    immutable set that contains the symbols for special tokens (literals, names and end-token).\
-    Has to be extended in a concrete parser adding all the symbols of the language.
     :cvar symbol_table: a dictionary that stores the token classes defined for the language.
     :type symbol_table: dict
     :cvar token_base_class: the base class for creating language's token classes.
     :type token_base_class: Token
     :cvar tokenizer: the language tokenizer compiled regexp.
     """
-    SYMBOLS: ClassVar[FrozenSet[str]] = SPECIAL_SYMBOLS
     token_base_class = Token
     tokenizer: Optional[Pattern[str]] = None
     symbol_table: MutableMapping[str, Type[TK_co]] = {}
 
     _start_token: TK_co
     source: str
-    tokens: Iterator[str]
+    tokens: Iterator[Match[str]]
     token: TK_co
     next_token: TK_co
     next_match: Optional[Match[str]]
@@ -440,10 +466,12 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         self._start_token = self.symbol_table['(start)'](self)
         self.token = self.next_token = self._start_token
 
+    def __repr__(self) -> str:
+        return f'{self.__class__.__name__}()'
+
     def __eq__(self, other: object) -> bool:
         return isinstance(other, Parser) and \
             self.token_base_class is other.token_base_class and \
-            self.SYMBOLS == other.SYMBOLS and \
             self.symbol_table == other.symbol_table
 
     def parse(self, source: str) -> TK_co:
@@ -457,7 +485,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         assert self.tokenizer, "Parser tokenizer is not built!"
         try:
             try:
-                self.tokens = iter(cast(Iterator[str], self.tokenizer.finditer(source)))
+                self.tokens = iter(self.tokenizer.finditer(source))
             except TypeError as err:
                 token = self.symbol_table['(invalid)'](self, source)
                 raise token.wrong_syntax('invalid source type, {}'.format(err))
@@ -472,71 +500,72 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
             self.next_match = None
             self.token = self.next_token = self._start_token
 
-    def advance(self, *symbols: str) -> TK_co:
+    def advance(self, *symbols: str, message: Optional[str] = None) -> TK_co:
         """
         The Pratt's function for advancing to next token.
 
         :param symbols: Optional arguments tuple. If not empty one of the provided \
         symbols is expected. If the next token's symbol differs the parser raises a \
         parse error.
-        :return: The next token instance.
+        :param message: Optional custom message for unexpected symbols.
+        :return: The current token instance.
         """
         value: Any
-        if self.next_token.symbol == '(end)' or \
-                symbols and self.next_token.symbol not in symbols:
+        if self.next_token.symbol == '(end)':
             raise self.next_token.wrong_syntax()
+        elif symbols and self.next_token.symbol not in symbols:
+            raise self.next_token.wrong_syntax(message)
 
         self.token = self.next_token
-        while True:
-            try:
-                self.next_match = cast(Match[str], next(self.tokens))
-            except StopIteration:
-                self.next_token = self.symbol_table['(end)'](self)
+
+        for self.next_match in self.tokens:
+            assert self.next_match is not None
+            if not self.next_match.group().isspace():
                 break
+        else:
+            self.next_token = self.symbol_table['(end)'](self)
+            return self.token
+
+        literal, symbol, name, unknown = self.next_match.groups()
+        if symbol is not None:
+            try:
+                self.next_token = self.symbol_table[symbol](self)
+            except KeyError:
+                if self.name_pattern.match(symbol) is None:
+                    self.next_token = self.symbol_table['(unknown)'](self, symbol)
+                    raise self.next_token.wrong_syntax()
+                self.next_token = self.symbol_table['(name)'](self, symbol)
+        elif literal is not None:
+            if literal[0] in '\'"':
+                value = self.unescape(literal)
+                self.next_token = self.symbol_table['(string)'](self, value)
+            elif 'e' in literal or 'E' in literal:
+                try:
+                    value = float(literal)
+                except ValueError as err:
+                    self.next_token = self.symbol_table['(invalid)'](self, literal)
+                    raise self.next_token.wrong_syntax(message=str(err))
+                else:
+                    self.next_token = self.symbol_table['(float)'](self, value)
+            elif '.' in literal:
+                try:
+                    value = Decimal(literal)
+                except DecimalException as err:
+                    self.next_token = self.symbol_table['(invalid)'](self, literal)
+                    raise self.next_token.wrong_syntax(message=str(err))
+                else:
+                    self.next_token = self.symbol_table['(decimal)'](self, value)
             else:
-                literal, symbol, name, unknown = self.next_match.groups()
-                if symbol is not None:
-                    try:
-                        self.next_token = self.symbol_table[symbol](self)
-                    except KeyError:
-                        if self.name_pattern.match(symbol) is None:
-                            self.next_token = self.symbol_table['(unknown)'](self, symbol)
-                            raise self.next_token.wrong_syntax()
-                        self.next_token = self.symbol_table['(name)'](self, symbol)
-                    break
-                elif literal is not None:
-                    if literal[0] in '\'"':
-                        value = self.unescape(literal)
-                        self.next_token = self.symbol_table['(string)'](self, value)
-                    elif 'e' in literal or 'E' in literal:
-                        try:
-                            value = float(literal)
-                        except ValueError as err:
-                            self.next_token = self.symbol_table['(invalid)'](self, literal)
-                            raise self.next_token.wrong_syntax(message=str(err))
-                        else:
-                            self.next_token = self.symbol_table['(float)'](self, value)
-                    elif '.' in literal:
-                        try:
-                            value = Decimal(literal)
-                        except DecimalException as err:
-                            self.next_token = self.symbol_table['(invalid)'](self, literal)
-                            raise self.next_token.wrong_syntax(message=str(err))
-                        else:
-                            self.next_token = self.symbol_table['(decimal)'](self, value)
-                    else:
-                        self.next_token = self.symbol_table['(integer)'](self, int(literal))
-                    break
-                elif name is not None:
-                    self.next_token = self.symbol_table['(name)'](self, name)
-                    break
-                elif unknown is not None:
-                    self.next_token = self.symbol_table['(unknown)'](self, unknown)
-                    break
-                elif str(self.next_match.group()).strip():
-                    msg = "unexpected matching %r: incompatible tokenizer"
-                    raise RuntimeError(msg % self.next_match.group())
-        return self.next_token
+                self.next_token = self.symbol_table['(integer)'](self, int(literal))
+        elif name is not None:
+            self.next_token = self.symbol_table['(name)'](self, name)
+        elif unknown is not None:
+            self.next_token = self.symbol_table['(unknown)'](self, unknown)
+        else:
+            msg = "unexpected matching %r: incompatible tokenizer"
+            raise RuntimeError(msg % self.next_match.group())
+
+        return self.token
 
     def advance_until(self, *stop_symbols: str) -> str:
         """
@@ -557,7 +586,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         source_chunk: List[str] = []
         while True:
             try:
-                self.next_match = cast(Match[str], next(self.tokens))
+                self.next_match = next(self.tokens)
             except StopIteration:
                 self.next_token = self.symbol_table['(end)'](self)
                 break
@@ -587,13 +616,11 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         :param rbp: right binding power for the expression.
         :return: left token.
         """
-        token = self.next_token
         self.advance()
-        left = token.nud()
+        left = self.token.nud()
         while rbp < self.next_token.lbp:
-            token = self.next_token
             self.advance()
-            left = token.led(left)
+            left = self.token.led(left)
         return cast(TK_co, left)
 
     @property
@@ -606,20 +633,14 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         Returns `True` if the parser is positioned at the start
         of the source, ignoring the spaces.
         """
-        return not bool(self.source[0:self.token.span[0]].strip())
+        return self.token.is_source_start()
 
     def is_line_start(self) -> bool:
         """
         Returns `True` if the parser is positioned at the start
         of a source line, ignoring the spaces.
         """
-        token_index = self.token.span[0]
-        try:
-            line_start = self.source[:token_index].rindex('\n') + 1
-        except ValueError:
-            return not bool(self.source[:token_index].strip())
-        else:
-            return not bool(self.source[line_start:token_index].strip())
+        return self.token.is_line_start()
 
     def is_spaced(self, before: bool = True, after: bool = True) -> bool:
         """
@@ -631,13 +652,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         :param after: if `True` considers also the extra spaces after \
         the current token symbol.
         """
-        start, end = self.token.span
-        try:
-            if before and start > 0 and self.source[start - 1] in ' \t\n':
-                return True
-            return after and self.source[end] in ' \t\n'
-        except IndexError:
-            return False
+        return self.token.is_spaced(before, after)
 
     @staticmethod
     def unescape(string_literal: str) -> str:
@@ -656,23 +671,27 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
             if ' ' in symbol:
                 raise ValueError("%r: a symbol can't contain whitespaces" % symbol)
 
+            lookup_name = kwargs.get('lookup_name', symbol)
             try:
-                token_class = cls.symbol_table[symbol]
+                token_class = cls.symbol_table[lookup_name]
             except KeyError:
-                # Register a new symbol and create a new custom class. The new class
-                # name is registered at parser class's module level.
-                if symbol not in cls.SYMBOLS:
-                    if symbol != '(start)':  # for backward compatibility
-                        raise NameError('%r is not a symbol of the parser %r.' % (symbol, cls))
+                # Register a new symbol and create a new custom class. The new token
+                # class is registered globally in the module of the parser class.
 
                 kwargs['symbol'] = symbol
+                kwargs['lookup_name'] = lookup_name
                 label = kwargs.get('label', 'symbol')
                 if isinstance(label, tuple):
                     label = kwargs['label'] = MultiLabel(*label)
 
-                token_class_name = "_{}{}".format(
-                    _symbol_to_classname(symbol), str(label).title().replace(' ', '')
-                )
+                if 'class_name' in kwargs:
+                    token_class_name = kwargs.pop('class_name')
+                else:
+                    token_class_name = "_{}{}".format(
+                        _symbol_to_classname(symbol),
+                        str(label).title().replace(' ', '')
+                    )
+
                 token_class_bases = kwargs.get('bases', (cls.token_base_class,))
                 kwargs.update({
                     '__module__': cls.__module__,
@@ -682,7 +701,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
                 token_class = cast(
                     Type[TK_co], ABCMeta(token_class_name, token_class_bases, kwargs)
                 )
-                cls.symbol_table[symbol] = token_class
+                cls.symbol_table[lookup_name] = token_class
                 MutableSequence.register(token_class)
                 setattr(sys.modules[cls.__module__], token_class_name, token_class)
 
@@ -690,7 +709,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
             raise TypeError("A string or a %r subclass requested, not %r." % (Token, symbol))
         else:
             token_class = symbol
-            if cls.symbol_table.get(symbol.symbol) is not token_class:
+            if cls.symbol_table.get(symbol.lookup_name) is not token_class:
                 raise ValueError("Token class %r is not registered." % token_class)
 
         for key, value in kwargs.items():
@@ -782,7 +801,7 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
         def bind(func: Callable[..., Any]) -> Callable[..., Any]:
             method_name = func.__name__.partition('_')[0]
             if not callable(getattr(token_class, method_name)):
-                raise TypeError(f"The attribute {method_name!r} is not a callable of {token_class}")
+                raise TypeError(f"{method_name!r} is not a method of {token_class}")
             setattr(token_class, method_name, func)
             return func
         return bind
@@ -791,19 +810,20 @@ class Parser(Generic[TK_co], metaclass=ParserMeta):
     def build(cls) -> None:
         """
         Builds the parser class. Checks if all declared symbols are defined
-        and builds a the regex tokenizer using the symbol related patterns.
+        and builds the regex tokenizer using the symbol related patterns.
         """
-        # For backward compatibility with external defined parsers
+        # Register a minimal set of special tokens
         if '(start)' not in cls.symbol_table:
             cls.register('(start)')
+        if '(end)' not in cls.symbol_table:
+            cls.register('(end)')
+        if '(invalid)' not in cls.symbol_table:
+            cls.register('(invalid)')
+        if '(unknown)' not in cls.symbol_table:
+            cls.register('(unknown)')
 
-        if not cls.SYMBOLS.issubset(cls.symbol_table.keys()):
-            unregistered = [s for s in cls.SYMBOLS if s not in cls.symbol_table]
-            raise ValueError("The parser %r has unregistered symbols: %r" % (cls, unregistered))
         cls.tokenizer = cls.create_tokenizer(cls.symbol_table)
 
-    build_tokenizer = build  # For backward compatibility
-
     @classmethod
     def create_tokenizer(cls, symbol_table: MutableMapping[str, Type[TK_co]]) -> Pattern[str]:
         """
