@@ -11,7 +11,7 @@ import calendar
 import datetime
 import decimal
 import re
-from typing import Iterator, List, Optional, Tuple, Union
+from typing import Any, Iterator, List, Optional, Tuple, Union
 from unicodedata import category
 
 from ..exceptions import xpath_error
@@ -34,6 +34,7 @@ def decimal_to_string(value: decimal.Decimal) -> str:
     Convert a Decimal value to a string representation
     that not includes exponent and with its decimals.
     """
+    exponent: Any
     sign, digits, exponent = value.as_tuple()
 
     if not exponent:
@@ -189,9 +190,11 @@ def format_digits(digits: str,
         else:
             separator = _separator.pop()
             chunks = fmt.split(separator)
-            repeat = len(chunks[-1])
-            if all(len(item) == repeat for item in chunks[1:-1]):
-                repeat += 1
+
+            if len(chunks[0]) > len(chunks[-1]):
+                repeat = None
+            elif all(len(item) == len(chunks[-1]) for item in chunks[1:-1]):
+                repeat = len(chunks[-1]) + 1
             else:
                 repeat = None
 
